// Minecraft Standalone Client
// Copyright (C) 2009-2010 Colin Bartolome
// Licensed under the GPL. See LICENSE.txt for details.

import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;

public class ServerDialog extends JDialog
{
   private static final String[] PARAMETERS = new String[]
   {
      "username",
      "sessionid",
      "loadmap_user",
      "loadmap_id",
      "server",
      "port",
      "mppass"
   };
   
   private static class BookmarkListModel extends AbstractListModel
   {
      private List<String> bookmarks;
      
      BookmarkListModel(Map<String, String> bookmarks)
      {
         this.bookmarks = new ArrayList<String>(bookmarks.keySet());
      }

      @Override
      public Object getElementAt(int index)
      {
         return bookmarks.get(index);
      }

      @Override
      public int getSize()
      {
         return bookmarks.size();
      }

      public void add(String name)
      {
         bookmarks.add(name);
         Collections.sort(bookmarks);
         fireContentsChanged(this, 0, getSize());
      }

      public void remove(int index)
      {
         bookmarks.remove(index);
         fireContentsChanged(this, 0, getSize());
      }
   }
   
   public ServerDialog(final Main main, final MainApplet applet)
   {
      super(main, "Select a Server", true);
      
      final Map<String, String> maps;
      final Map<String, String> servers;
      final Map<String, String> bookmarks;
      
      try
      {
         maps = parseMaps(getPage("http://www.minecraft.net/", main.getLoginCookie()));
         servers = parseServers(getPage("http://www.minecraft.net/servers.jsp", main.getLoginCookie()));
         bookmarks = Main.getBookmarks();
      }
      catch (IOException ioe)
      {
         JOptionPane.showMessageDialog(this, "Could not retrieve server list. Going single-player.");
         
         return;
      }
      
      final Map<String, Map<String, String>> lists = new HashMap<String, Map<String,String>>();
      
      lists.put("maps", maps);
      lists.put("servers", servers);
      lists.put("bookmarks", bookmarks);
      
      JPanel panel = new JPanel();
      
      panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
      panel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
      
      final JList mapList;
      final JList serverList;
      final JList bookmarkList;
      final BookmarkListModel bookmarkListModel;
      final JTextField textField = new JTextField(50);
      Box box;
      final JButton addButton;
      JButton editButton;
      JButton deleteButton;
      JButton goButton;
      
      box = Box.createHorizontalBox();
      box.add(new JLabel("Your Maps:"));
      box.add(Box.createHorizontalGlue());
      panel.add(box);

      panel.add(Box.createVerticalStrut(5));

      panel.add(new JScrollPane(mapList = new JList(maps.keySet().toArray())));
      mapList.setName("maps");
      mapList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(new JLabel("Servers:"));
      box.add(Box.createHorizontalGlue());
      panel.add(box);

      panel.add(Box.createVerticalStrut(5));

      panel.add(new JScrollPane(serverList = new JList(servers.keySet().toArray())));
      serverList.setName("servers");
      serverList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(new JLabel("Bookmarks:"));
      box.add(Box.createHorizontalGlue());
      box.add(addButton = new JButton("Add"));
      box.add(Box.createHorizontalStrut(5));
      box.add(editButton = new JButton("Edit"));
      box.add(Box.createHorizontalStrut(5));
      box.add(deleteButton = new JButton("Delete"));
      panel.add(box);
      
      panel.add(Box.createVerticalStrut(5));

      panel.add(new JScrollPane(bookmarkList = new JList(bookmarkListModel = new BookmarkListModel(bookmarks))));
      bookmarkList.setName("bookmarks");
      bookmarkList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(new JLabel("<html>Choose a server above or paste a <code>play.jsp</code> URL below. Leave blank to generate a level.</html>"));
      box.add(Box.createHorizontalGlue());
      panel.add(box);
      
      panel.add(Box.createVerticalStrut(5));
      
      box = Box.createHorizontalBox();
      box.add(new JLabel("URL:"));
      box.add(Box.createHorizontalStrut(5));
      box.add(textField);
      panel.add(box);
      
      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(Box.createHorizontalGlue());
      box.add(goButton = new JButton("Go!"));
      box.add(Box.createHorizontalGlue());
      panel.add(box);
      
      addButton.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent evt)
         {
            String name;
            String url;
            
            do
            {
               name = JOptionPane.showInputDialog(addButton, "Enter a name for the bookmark:");
               
               if (name != null && bookmarks.containsKey(name))
                  JOptionPane.showMessageDialog(addButton, "That name is already in use. Please choose another.");
            } while (name != null && !name.isEmpty() && bookmarks.containsKey(name));
            
            if (name == null || name.isEmpty())
               return;
            
            do
            {
               url = JOptionPane.showInputDialog(addButton, "Enter the URL for the bookmark:");
            } while (url != null && url.isEmpty());
            
            if (url == null)
               return;
            
            bookmarks.put(name, url);
            bookmarkListModel.add(name);
            bookmarkList.clearSelection();
            Main.saveBookmark(name, url);
         }
      });
      
      editButton.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent evt)
         {
            int index = bookmarkList.getSelectedIndex();
            
            if (index < 0 || index >= bookmarkListModel.getSize())
               return;
            
            String name = (String)bookmarkListModel.getElementAt(index);
            String url = bookmarks.get(name);
            
            do
            {
               url = JOptionPane.showInputDialog(addButton, "Enter the URL for the bookmark:", url);
            } while (url != null && url.isEmpty());

            if (url == null)
               return;
            
            bookmarks.put(name, url);
            bookmarkList.clearSelection();
            Main.saveBookmark(name, url);
         }
      });
      
      deleteButton.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent evt)
         {
            int index = bookmarkList.getSelectedIndex();
            
            if (index < 0 || index >= bookmarkListModel.getSize())
               return;
            
            String name = (String)bookmarkListModel.getElementAt(index);
            
            bookmarks.remove(name);
            bookmarkListModel.remove(index);
            bookmarkList.clearSelection();
            Main.saveBookmark(name, null);
         }
      });

      goButton.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent evt)
         {
            goForIt(main, applet, textField.getText());
         }
      });
      
      ListSelectionListener listSelectionListener = new ListSelectionListener()
      {
         @Override
         public void valueChanged(ListSelectionEvent evt)
         {
            JList source = (JList)evt.getSource();
            Object value = source.getSelectedValue();
            
            if (value == null)
               return;
            
            textField.setText(lists.get(source.getName()).get(value));
            
            if (!source.equals(mapList))
               mapList.clearSelection();
            if (!source.equals(serverList))
               serverList.clearSelection();
            if (!source.equals(bookmarkList))
               bookmarkList.clearSelection();
         }
      };
      
      MouseListener mouseListener = new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent evt)
         {
            if (evt.getClickCount() == 2)
               goForIt(main, applet, textField.getText());
         }
      };
      
      mapList.addListSelectionListener(listSelectionListener);
      mapList.addMouseListener(mouseListener);

      serverList.addListSelectionListener(listSelectionListener);
      serverList.addMouseListener(mouseListener);
      
      bookmarkList.addListSelectionListener(listSelectionListener);
      bookmarkList.addMouseListener(mouseListener);
      
      getRootPane().setDefaultButton(goButton);
      
      add(panel);
      
      pack();
      setLocationRelativeTo(null);
      setVisible(true);
   }
   
   private void goForIt(Main main, MainApplet applet, String text)
   {
      try
      {
         String url = text.trim();
         
         if (url.isEmpty())
            url = "http://www.minecraft.net/play.jsp";
         
         parseParameters(applet, getPage(url, main.getLoginCookie()));
      
         setVisible(false);
      }
      catch (IOException ioe)
      {
         JOptionPane.showMessageDialog(ServerDialog.this, "Could not parse the necessary parameters. Is the URL right?");
      }
   }

   private String getPage(String url, String loginCookie) throws IOException
   {
      StringBuilder rtn = new StringBuilder();
      URLConnection connection = new URL(url).openConnection();
      
      connection.addRequestProperty("Cookie", "_lgn=" + loginCookie);
      connection.setDoInput(true);
      connection.connect();
      
      BufferedReader in = new BufferedReader(new InputStreamReader(connection.getInputStream()));
      char[] buffer = new char[4096];
      int charsRead;
      
      while ((charsRead = in.read(buffer)) != -1)
         rtn.append(buffer, 0, charsRead);
      
      in.close();
      
      return rtn.toString();
   }
   
   private String decodeServerName(String serverName)
   {
      String rtn = serverName;
      
      rtn = rtn.replaceAll("&lt;", "<");
      rtn = rtn.replaceAll("&gt;", ">");
      
      return rtn;
   }
   
   private Map<String, String> parseMaps(String text)
   {
      Map<String, String> servers = new TreeMap<String, String>();
      int index = text.indexOf("<a href=\"");
      String serverName;
      String serverURL;
      
      while (index != -1)
      {
         index = text.indexOf("play.jsp?user=", index);
         
         if (index == -1)
            break;
         
         serverURL = text.substring(index, text.indexOf("\"", index));
         index = text.indexOf(">", index) + ">".length();
         serverName = text.substring(index, text.indexOf("</a>", index));
         
         servers.put(decodeServerName(serverName), "http://www.minecraft.net/" + serverURL);
      }
      
      return servers;
   }
   
   private Map<String, String> parseServers(String text)
   {
      Map<String, String> servers = new TreeMap<String, String>();
      int index = text.indexOf("<a href=\"");
      String serverName;
      String serverURL;
      
      while (index != -1)
      {
         index = text.indexOf("play.jsp?server=", index);
         
         if (index == -1)
            break;
         
         serverURL = text.substring(index, text.indexOf("\"", index));
         index = text.indexOf("<b>", index) + "<b>".length();
         serverName = text.substring(index, text.indexOf("</b>", index));
         index = text.indexOf("</a>", index) + "</a>".length();
         serverName += text.substring(index, text.indexOf('\n', index));
         
         servers.put(decodeServerName(serverName), "http://www.minecraft.net/" + serverURL);
      }
      
      return servers;
   }
   
   private void parseParameters(MainApplet applet, String text)
   {
      for (String parameter : PARAMETERS)
      {
         int index = text.indexOf("name=\"" + parameter + "\"");
         
         if (index != -1)
         {
            index = text.indexOf("value=\"", index) + "value=\"".length();
            
            String value = text.substring(index, text.indexOf("\"", index));
            
            System.out.println(parameter + ": " + value);
            
            applet.putParameter(parameter, value);
         }
      }
   }
}
